/*
Copyright 2021 The KCP Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package workspace

import (
	"context"
	"fmt"
	"math/rand"
	"sort"
	"testing"
	"time"

	"github.com/google/go-cmp/cmp"
	"github.com/kcp-dev/logicalcluster"
	"github.com/stretchr/testify/require"

	apiextensionsclientset "k8s.io/apiextensions-apiserver/pkg/client/clientset/clientset"
	"k8s.io/apimachinery/pkg/api/errors"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/types"
	"k8s.io/apimachinery/pkg/util/sets"
	"k8s.io/apimachinery/pkg/util/wait"
	"k8s.io/apiserver/pkg/endpoints/discovery"
	clientgodiscovery "k8s.io/client-go/discovery"
	kubernetesclientset "k8s.io/client-go/kubernetes"
	"k8s.io/client-go/rest"
	"k8s.io/client-go/tools/clientcmd"
	"k8s.io/klog/v2"

	apisv1alpha1 "github.com/kcp-dev/kcp/pkg/apis/apis/v1alpha1"
	"github.com/kcp-dev/kcp/pkg/apis/third_party/conditions/util/conditions"
	kcpclient "github.com/kcp-dev/kcp/pkg/client/clientset/versioned"
	kubefixtures "github.com/kcp-dev/kcp/test/e2e/fixtures/kube"
	fixturewildwest "github.com/kcp-dev/kcp/test/e2e/fixtures/wildwest"
	"github.com/kcp-dev/kcp/test/e2e/fixtures/wildwest/apis/wildwest"
	"github.com/kcp-dev/kcp/test/e2e/fixtures/wildwest/apis/wildwest/v1alpha1"
	wildwestclientset "github.com/kcp-dev/kcp/test/e2e/fixtures/wildwest/client/clientset/versioned"
	"github.com/kcp-dev/kcp/test/e2e/framework"
)

func deploymentsAPIResourceList(workspaceName string) *metav1.APIResourceList {
	return &metav1.APIResourceList{
		TypeMeta: metav1.TypeMeta{
			Kind:       "APIResourceList",
			APIVersion: "v1",
		},
		GroupVersion: "apps/v1",
		APIResources: []metav1.APIResource{
			{
				Kind:               "Deployment",
				Name:               "deployments",
				SingularName:       "deployment",
				Namespaced:         true,
				Verbs:              metav1.Verbs{"get", "list", "patch", "update", "watch"},
				StorageVersionHash: discovery.StorageVersionHash(workspaceName, "apps", "v1", "Deployment"),
				Categories:         []string{"all"},
				ShortNames:         []string{"deploy"},
			},
			{
				Kind:               "Deployment",
				Name:               "deployments/status",
				Namespaced:         true,
				Verbs:              metav1.Verbs{"get", "patch", "update"},
				StorageVersionHash: "",
			},
		},
	}
}

func requiredCoreAPIResourceList(workspaceName string) *metav1.APIResourceList {
	return &metav1.APIResourceList{
		TypeMeta: metav1.TypeMeta{
			Kind: "APIResourceList",
		},
		GroupVersion: "v1",
		APIResources: []metav1.APIResource{
			{
				Kind:               "ConfigMap",
				Name:               "configmaps",
				SingularName:       "configmap",
				Namespaced:         true,
				Verbs:              metav1.Verbs{"get", "list", "patch", "update", "watch"},
				StorageVersionHash: discovery.StorageVersionHash(workspaceName, "", "v1", "ConfigMap"),
			},
			{
				Kind:               "Namespace",
				Name:               "namespaces",
				SingularName:       "namespace",
				Namespaced:         false,
				Verbs:              metav1.Verbs{"get", "list", "patch", "update", "watch"},
				StorageVersionHash: discovery.StorageVersionHash(workspaceName, "", "v1", "Namespace"),
			},
			{
				Kind:               "Namespace",
				Name:               "namespaces/status",
				SingularName:       "",
				Namespaced:         false,
				Verbs:              metav1.Verbs{"get", "patch", "update"},
				StorageVersionHash: "",
			},
			{
				Kind:               "Secret",
				Name:               "secrets",
				SingularName:       "secret",
				Namespaced:         true,
				Verbs:              metav1.Verbs{"get", "list", "patch", "update", "watch"},
				StorageVersionHash: discovery.StorageVersionHash(workspaceName, "", "v1", "Secret"),
			},
			{
				Kind:               "ServiceAccount",
				Name:               "serviceaccounts",
				SingularName:       "serviceaccount",
				Namespaced:         true,
				Verbs:              metav1.Verbs{"get", "list", "patch", "update", "watch"},
				StorageVersionHash: discovery.StorageVersionHash(workspaceName, "", "v1", "ServiceAccount"),
			},
		},
	}
}

func addToAPIResourceList(list *metav1.APIResourceList, resources ...metav1.APIResource) *metav1.APIResourceList {
	list.APIResources = append(list.APIResources, resources...)

	return list
}

func TestSyncerVirtualWorkspace(t *testing.T) {
	t.Parallel()

	server := framework.SharedKcpServer(t)
	orgClusterName := framework.NewOrganizationFixture(t, server)

	var testCases = []struct {
		name string
		work func(t *testing.T, kubelikeSyncerVWConfig, wildwestSyncerVWConfig *rest.Config, kubelikeClusterName, wildwestClusterName logicalcluster.Name)
	}{
		{
			name: "isolated API domains per syncer",
			work: func(t *testing.T, kubelikeSyncerVWConfig, wildwestSyncerVWConfig *rest.Config, kubelikeClusterName, wildwestClusterName logicalcluster.Name) {
				kubelikeVWDiscoverClusterClient, err := clientgodiscovery.NewDiscoveryClientForConfig(kubelikeSyncerVWConfig)
				require.NoError(t, err)

				t.Logf("Check discovery in kubelike virtual workspace")
				_, kubelikeAPIResourceLists, err := kubelikeVWDiscoverClusterClient.WithCluster(logicalcluster.Wildcard).ServerGroupsAndResources()
				require.NoError(t, err)
				require.Empty(t, cmp.Diff([]*metav1.APIResourceList{
					deploymentsAPIResourceList(kubelikeClusterName.String()),
					{
						TypeMeta: metav1.TypeMeta{
							Kind:       "APIResourceList",
							APIVersion: "v1",
						},
						GroupVersion: "networking.k8s.io/v1",
						APIResources: []metav1.APIResource{
							{
								Kind:               "Ingress",
								Name:               "ingresses",
								SingularName:       "ingress",
								Namespaced:         true,
								Verbs:              metav1.Verbs{"get", "list", "patch", "update", "watch"},
								ShortNames:         []string{"ing"},
								StorageVersionHash: discovery.StorageVersionHash(kubelikeClusterName.String(), "networking.k8s.io", "v1", "Ingress"),
							},
							{
								Kind:               "Ingress",
								Name:               "ingresses/status",
								Namespaced:         true,
								Verbs:              metav1.Verbs{"get", "patch", "update"},
								StorageVersionHash: "",
							},
						},
					},
					addToAPIResourceList(
						requiredCoreAPIResourceList(kubelikeClusterName.String()),
						metav1.APIResource{
							Kind:               "Service",
							Name:               "services",
							SingularName:       "service",
							Namespaced:         true,
							Verbs:              metav1.Verbs{"get", "list", "patch", "update", "watch"},
							StorageVersionHash: discovery.StorageVersionHash(kubelikeClusterName.String(), "", "v1", "Service"),
							Categories:         []string{"all"},
							ShortNames:         []string{"svc"},
						},
						metav1.APIResource{
							Kind:               "Service",
							Name:               "services/status",
							SingularName:       "",
							Namespaced:         true,
							Verbs:              metav1.Verbs{"get", "patch", "update"},
							StorageVersionHash: "",
						}),
				}, sortAPIResourceList(kubelikeAPIResourceLists)))

				t.Logf("Check discovery in wildwest virtual workspace")
				wildwestVWDiscoverClusterClient, err := clientgodiscovery.NewDiscoveryClientForConfig(wildwestSyncerVWConfig)
				require.NoError(t, err)
				_, wildwestAPIResourceLists, err := wildwestVWDiscoverClusterClient.WithCluster(logicalcluster.Wildcard).ServerGroupsAndResources()
				require.NoError(t, err)
				require.Empty(t, cmp.Diff([]*metav1.APIResourceList{
					deploymentsAPIResourceList(wildwestClusterName.String()),
					requiredCoreAPIResourceList(wildwestClusterName.String()),
					{
						TypeMeta: metav1.TypeMeta{
							Kind:       "APIResourceList",
							APIVersion: "v1",
						},
						GroupVersion: "wildwest.dev/v1alpha1",
						APIResources: []metav1.APIResource{
							{
								Kind:               "Cowboy",
								Name:               "cowboys",
								SingularName:       "cowboy",
								Namespaced:         true,
								Verbs:              metav1.Verbs{"get", "list", "patch", "update", "watch"},
								StorageVersionHash: discovery.StorageVersionHash(wildwestClusterName.String(), "wildwest.dev", "v1alpha1", "Cowboy"),
							},
							{
								Kind:               "Cowboy",
								Name:               "cowboys/status",
								Namespaced:         true,
								Verbs:              metav1.Verbs{"get", "patch", "update"},
								StorageVersionHash: "",
							},
						},
					},
				}, sortAPIResourceList(wildwestAPIResourceLists)))
			},
		},
		{
			name: "access kcp resources through syncer virtual workspace",
			work: func(t *testing.T, kubelikeSyncerVWConfig, wildwestSyncerVWConfig *rest.Config, kubelikeClusterName, wildwestClusterName logicalcluster.Name) {
				ctx, cancelFunc := context.WithCancel(context.Background())
				t.Cleanup(cancelFunc)

				wildwestClusterClient, err := wildwestclientset.NewClusterForConfig(server.DefaultConfig(t))
				require.NoError(t, err)

				t.Log("Create cowboy luckyluke")
				_, err = wildwestClusterClient.Cluster(wildwestClusterName).WildwestV1alpha1().Cowboys("default").Create(ctx, &v1alpha1.Cowboy{
					ObjectMeta: metav1.ObjectMeta{
						Name: "luckyluke",
					},
					Spec: v1alpha1.CowboySpec{
						Intent: "should catch joe",
					},
				}, metav1.CreateOptions{})
				require.NoError(t, err)

				vwClusterClient, err := wildwestclientset.NewClusterForConfig(wildwestSyncerVWConfig)
				require.NoError(t, err)

				t.Log("Verify there is one cowboy via direct access")
				kcpCowboys, err := wildwestClusterClient.Cluster(wildwestClusterName).WildwestV1alpha1().Cowboys("").List(ctx, metav1.ListOptions{})
				require.NoError(t, err)
				require.Len(t, kcpCowboys.Items, 1)

				t.Log("Wait until the virtual workspace has the resource")
				require.Eventually(t, func() bool {
					// resources show up asynchronously, so we have to try until List works. Then it should return all object immediately.
					_, err := vwClusterClient.Cluster(logicalcluster.Wildcard).WildwestV1alpha1().Cowboys("").List(ctx, metav1.ListOptions{})
					return err == nil
				}, wait.ForeverTestTimeout, time.Millisecond*100)

				t.Log("Verify there is one cowboy via virtual workspace wildcard request")
				virtualWorkspaceCowboys, err := vwClusterClient.Cluster(logicalcluster.Wildcard).WildwestV1alpha1().Cowboys("").List(ctx, metav1.ListOptions{})
				require.NoError(t, err)
				for _, cs := range virtualWorkspaceCowboys.Items {
					if logicalcluster.From(&cs) == wildwestClusterName {
						require.Equal(t, "luckyluke", cs.Name)
					}
				}

				t.Log("Verify there is luckyluke via virtual workspace workspace request")
				kcpCowboy, err := wildwestClusterClient.Cluster(wildwestClusterName).WildwestV1alpha1().Cowboys("default").Get(ctx, "luckyluke", metav1.GetOptions{})
				require.NoError(t, err)
				virtualWorkspaceCowboy, err := vwClusterClient.Cluster(wildwestClusterName).WildwestV1alpha1().Cowboys("default").Get(ctx, "luckyluke", metav1.GetOptions{})
				require.NoError(t, err)
				require.Equal(t, kcpCowboy.UID, virtualWorkspaceCowboy.UID)
				require.Equal(t, kcpCowboy.Spec, virtualWorkspaceCowboy.Spec)
				require.Equal(t, kcpCowboy.Status, virtualWorkspaceCowboy.Status)

				t.Log("Patch luckyluke via virtual workspace to report in status that joe is in prison")
				_, err = vwClusterClient.Cluster(wildwestClusterName).WildwestV1alpha1().Cowboys("default").Patch(ctx, "luckyluke", types.MergePatchType, []byte("{\"status\":{\"result\":\"joe in prison\"}}"), metav1.PatchOptions{}, "status")
				require.NoError(t, err)

				t.Log("Patch luckyluke via virtual workspace to catch averell")
				_, err = vwClusterClient.Cluster(wildwestClusterName).WildwestV1alpha1().Cowboys("default").Patch(ctx, "luckyluke", types.MergePatchType, []byte("{\"spec\":{\"intent\":\"should catch averell\"}}"), metav1.PatchOptions{})
				require.NoError(t, err)

				t.Log("Verify that luckyluke has both spec and status changed")
				modifiedkcpCowboy, err := wildwestClusterClient.Cluster(wildwestClusterName).WildwestV1alpha1().Cowboys("default").Get(ctx, "luckyluke", metav1.GetOptions{})
				require.NoError(t, err)
				require.NotEqual(t, kcpCowboy.ResourceVersion, modifiedkcpCowboy)

				t.Log("Verify resource version, managed fields and generation")
				expectedModifiedKcpCowboy := kcpCowboy.DeepCopy()
				expectedModifiedKcpCowboy.Status.Result = "joe in prison"
				expectedModifiedKcpCowboy.Spec.Intent = "should catch averell"
				require.Empty(t, cmp.Diff(expectedModifiedKcpCowboy.Spec, modifiedkcpCowboy.Spec))
				require.Empty(t, cmp.Diff(expectedModifiedKcpCowboy.Status, modifiedkcpCowboy.Status))
			},
		},
		{
			name: "access kcp resources through syncer virtual workspace, from a other workspace to the wildwest resources through an APIBinding",
			work: func(t *testing.T, kubelikeSyncerVWConfig, wildwestSyncerVWConfig *rest.Config, kubelikeClusterName, wildwestClusterName logicalcluster.Name) {
				ctx, cancelFunc := context.WithCancel(context.Background())
				t.Cleanup(cancelFunc)

				kcpClusterClient, err := kcpclient.NewClusterForConfig(server.DefaultConfig(t))
				require.NoError(t, err)

				otherWorkspace := framework.NewWorkspaceFixture(t, server, orgClusterName, "Universal")

				t.Logf("Create a binding in the other workspace")
				binding := &apisv1alpha1.APIBinding{
					ObjectMeta: metav1.ObjectMeta{
						Name: "kubernetes",
					},
					Spec: apisv1alpha1.APIBindingSpec{
						Reference: apisv1alpha1.ExportReference{
							Workspace: &apisv1alpha1.WorkspaceExportReference{
								Path:       wildwestClusterName.String(),
								ExportName: "kubernetes",
							},
						},
					},
				}
				_, err = kcpClusterClient.Cluster(otherWorkspace).ApisV1alpha1().APIBindings().Create(ctx, binding, metav1.CreateOptions{})
				require.NoError(t, err)

				t.Logf("Wait for binding to be ready")
				require.Eventually(t, func() bool {
					binding, err := kcpClusterClient.Cluster(otherWorkspace).ApisV1alpha1().APIBindings().Get(ctx, binding.Name, metav1.GetOptions{})
					if err != nil {
						klog.Errorf("Failed to list Locations: %v", err)
						return false
					}
					return conditions.IsTrue(binding, apisv1alpha1.InitialBindingCompleted)
				}, wait.ForeverTestTimeout, time.Millisecond*100)

				t.Logf("Wait for binding to have cowboy resource")
				require.Eventually(t, func() bool {
					binding, err := kcpClusterClient.Cluster(otherWorkspace).ApisV1alpha1().APIBindings().Get(ctx, binding.Name, metav1.GetOptions{})
					if err != nil {
						klog.Errorf("Failed to list Locations: %v", err)
						return false
					}
					for _, r := range binding.Status.BoundResources {
						if r.Resource == "cowboys" {
							return true
						}
					}
					return false
				}, wait.ForeverTestTimeout, time.Millisecond*100)

				wildwestClusterClient, err := wildwestclientset.NewClusterForConfig(server.DefaultConfig(t))
				require.NoError(t, err)

				t.Logf("Wait for being able to list cowboys in the other workspace (kubelike) through the virtual workspace")
				require.Eventually(t, func() bool {
					_, err := wildwestClusterClient.Cluster(otherWorkspace).WildwestV1alpha1().Cowboys("").List(ctx, metav1.ListOptions{})
					if errors.IsNotFound(err) {
						return false
					} else if err != nil {
						klog.Errorf("Failed to list Services: %v", err)
						return false
					}
					return true
				}, wait.ForeverTestTimeout, time.Millisecond*100)

				t.Log("Create cowboy luckyluke")
				_, err = wildwestClusterClient.Cluster(otherWorkspace).WildwestV1alpha1().Cowboys("default").Create(ctx, &v1alpha1.Cowboy{
					ObjectMeta: metav1.ObjectMeta{
						Name: "luckyluke",
					},
					Spec: v1alpha1.CowboySpec{
						Intent: "should catch joe",
					},
				}, metav1.CreateOptions{})
				require.NoError(t, err)

				vwClusterClient, err := wildwestclientset.NewClusterForConfig(wildwestSyncerVWConfig)
				require.NoError(t, err)

				t.Log("Verify there is one cowboy via direct access")
				kcpCowboys, err := wildwestClusterClient.Cluster(otherWorkspace).WildwestV1alpha1().Cowboys("").List(ctx, metav1.ListOptions{})
				require.NoError(t, err)
				require.Len(t, kcpCowboys.Items, 1)

				t.Log("Wait until the virtual workspace has the resource")
				require.Eventually(t, func() bool {
					// resources show up asynchronously, so we have to try until List works. Then it should return all object immediately.
					_, err := vwClusterClient.Cluster(logicalcluster.Wildcard).WildwestV1alpha1().Cowboys("").List(ctx, metav1.ListOptions{})
					return err == nil
				}, wait.ForeverTestTimeout, time.Millisecond*100)

				t.Log("Verify there is one cowboy via virtual workspace via wildcard request")
				virtualWorkspaceCowboys, err := vwClusterClient.Cluster(logicalcluster.Wildcard).WildwestV1alpha1().Cowboys("").List(ctx, metav1.ListOptions{})
				require.NoError(t, err)
				for _, cs := range virtualWorkspaceCowboys.Items {
					if logicalcluster.From(&cs) == otherWorkspace {
						require.Equal(t, "luckyluke", cs.Name)
					}
				}

				t.Log("Verify there is luckyluke via virtual workspace")
				kcpCowboy, err := wildwestClusterClient.Cluster(otherWorkspace).WildwestV1alpha1().Cowboys("default").Get(ctx, "luckyluke", metav1.GetOptions{})
				require.NoError(t, err)
				virtualWorkspaceCowboy, err := vwClusterClient.Cluster(otherWorkspace).WildwestV1alpha1().Cowboys("default").Get(ctx, "luckyluke", metav1.GetOptions{})
				require.NoError(t, err)
				require.Equal(t, kcpCowboy.UID, virtualWorkspaceCowboy.UID)
				require.Empty(t, cmp.Diff(kcpCowboy.Spec, virtualWorkspaceCowboy.Spec))
				require.Empty(t, cmp.Diff(kcpCowboy.Status, virtualWorkspaceCowboy.Status))

				t.Log("Patch luckyluke via virtual workspace to report in status that joe is in prison")
				_, err = vwClusterClient.Cluster(otherWorkspace).WildwestV1alpha1().Cowboys("default").Patch(ctx, "luckyluke", types.MergePatchType, []byte("{\"status\":{\"result\":\"joe in prison\"}}"), metav1.PatchOptions{}, "status")
				require.NoError(t, err)

				t.Log("Patch luckyluke via virtual workspace to catch averell")
				_, err = vwClusterClient.Cluster(otherWorkspace).WildwestV1alpha1().Cowboys("default").Patch(ctx, "luckyluke", types.MergePatchType, []byte("{\"spec\":{\"intent\":\"should catch averell\"}}"), metav1.PatchOptions{})
				require.NoError(t, err)

				t.Log("Verify that luckyluke has both spec and status changed")
				modifiedkcpCowboy, err := wildwestClusterClient.Cluster(otherWorkspace).WildwestV1alpha1().Cowboys("default").Get(ctx, "luckyluke", metav1.GetOptions{})
				require.NoError(t, err)
				require.NotEqual(t, kcpCowboy.ResourceVersion, modifiedkcpCowboy)

				t.Log("Verify resource version, managed fields and generation")
				expectedModifiedKcpCowboy := kcpCowboy.DeepCopy()
				expectedModifiedKcpCowboy.Status.Result = "joe in prison"
				expectedModifiedKcpCowboy.Spec.Intent = "should catch averell"
				require.Empty(t, cmp.Diff(expectedModifiedKcpCowboy.Spec, modifiedkcpCowboy.Spec))
				require.Empty(t, cmp.Diff(expectedModifiedKcpCowboy.Status, modifiedkcpCowboy.Status))
			},
		},
	}

	for i := range testCases {
		testCase := testCases[i]
		t.Run(testCase.name, func(t *testing.T) {
			t.Parallel()

			ctx, cancelFunc := context.WithCancel(context.Background())
			t.Cleanup(cancelFunc)

			kubeClusterClient, err := kubernetesclientset.NewClusterForConfig(server.DefaultConfig(t))
			require.NoError(t, err)
			wildwestClusterClient, err := wildwestclientset.NewClusterForConfig(server.DefaultConfig(t))
			require.NoError(t, err)

			kubelikeWorkspace := framework.NewWorkspaceFixture(t, server, orgClusterName, "Universal")

			t.Logf("Deploying syncer into workspace %s", kubelikeWorkspace)
			_ = framework.SyncerFixture{
				ResourcesToSync:      sets.NewString("ingresses.networking.k8s.io", "services"),
				UpstreamServer:       server,
				WorkspaceClusterName: kubelikeWorkspace,
				WorkloadClusterName:  "kubelike",
				InstallCRDs: func(config *rest.Config, isLogicalCluster bool) {
					if !isLogicalCluster {
						// Only need to install services and ingresses in a logical cluster
						return
					}
					sinkCrdClient, err := apiextensionsclientset.NewForConfig(config)
					require.NoError(t, err, "failed to create apiextensions client")
					t.Logf("Installing test CRDs into sink cluster...")
					kubefixtures.Create(t, sinkCrdClient.ApiextensionsV1().CustomResourceDefinitions(),
						metav1.GroupResource{Group: "core.k8s.io", Resource: "services"},
						metav1.GroupResource{Group: "networking.k8s.io", Resource: "ingresses"},
					)
					require.NoError(t, err)
				},
			}.Start(t)

			t.Log("Waiting for ingresses crd to be imported and available in the kubelike source cluster...")
			require.Eventually(t, func() bool {
				_, err := kubeClusterClient.Cluster(kubelikeWorkspace).NetworkingV1().Ingresses("").List(ctx, metav1.ListOptions{})
				if err != nil {
					t.Logf("error seen waiting for ingresses crd to become active: %v", err)
					return false
				}
				return true
			}, wait.ForeverTestTimeout, time.Millisecond*100)

			t.Log("Waiting for services crd to be imported and available in the kubelike source cluster...")
			require.Eventually(t, func() bool {
				_, err := kubeClusterClient.Cluster(kubelikeWorkspace).CoreV1().Services("").List(ctx, metav1.ListOptions{})
				if err != nil {
					t.Logf("error seen waiting for services crd to become active: %v", err)
					return false
				}
				return true
			}, wait.ForeverTestTimeout, time.Millisecond*100)

			t.Log("Setting up an unrelated workspace with cowboys...")

			unrelatedWorkspace := framework.NewWorkspaceFixture(t, server, orgClusterName, "Universal")
			unrelatedWorkspaceWorkspaceClient := wildwestClusterClient.Cluster(unrelatedWorkspace)

			sourceCrdClient, err := apiextensionsclientset.NewClusterForConfig(server.DefaultConfig(t))
			require.NoError(t, err)

			fixturewildwest.Create(t, sourceCrdClient.Cluster(unrelatedWorkspace).ApiextensionsV1().CustomResourceDefinitions(), metav1.GroupResource{Group: wildwest.GroupName, Resource: "cowboys"})

			t.Log("Waiting for cowboys crd to be imported and available in the kubelike workspace...")
			require.Eventually(t, func() bool {
				_, err := unrelatedWorkspaceWorkspaceClient.WildwestV1alpha1().Cowboys("").List(ctx, metav1.ListOptions{})
				if err != nil {
					t.Logf("error seen waiting for cowboys crd to become active: %v", err)
					return false
				}
				return true
			}, wait.ForeverTestTimeout, time.Millisecond*100)

			t.Log("Create cowboy unluckyluke")
			_, err = unrelatedWorkspaceWorkspaceClient.WildwestV1alpha1().Cowboys("default").Create(ctx, &v1alpha1.Cowboy{
				ObjectMeta: metav1.ObjectMeta{
					Name: "unluckyluke",
				},
				Spec: v1alpha1.CowboySpec{
					Intent: "will never catch joe",
				},
			}, metav1.CreateOptions{})
			require.NoError(t, err)

			wildwestWorkspace := framework.NewWorkspaceFixture(t, server, orgClusterName, "Universal")
			wildwestWorkloadClusterName := fmt.Sprintf("wildwest-%d", +rand.Intn(1000000))

			t.Logf("Deploying syncer into workspace %s", wildwestWorkspace)
			_ = framework.SyncerFixture{
				ResourcesToSync:      sets.NewString("cowboys.wildwest.dev"),
				UpstreamServer:       server,
				WorkspaceClusterName: wildwestWorkspace,
				WorkloadClusterName:  wildwestWorkloadClusterName,
				InstallCRDs: func(config *rest.Config, isLogicalCluster bool) {
					// Always install the crd regardless of whether the target is
					// logical or not since cowboys is not a native type.
					sinkCrdClient, err := apiextensionsclientset.NewForConfig(config)
					require.NoError(t, err)
					t.Log("Installing test CRDs into sink cluster...")
					fixturewildwest.Create(t, sinkCrdClient.ApiextensionsV1().CustomResourceDefinitions(), metav1.GroupResource{Group: wildwest.GroupName, Resource: "cowboys"})
				},
			}.Start(t)

			t.Log("Waiting for cowboys crd to be imported and available in the wildwest source workspace...")
			require.Eventually(t, func() bool {
				_, err := wildwestClusterClient.Cluster(wildwestWorkspace).WildwestV1alpha1().Cowboys("").List(ctx, metav1.ListOptions{})
				if err != nil {
					t.Logf("error seen waiting for cowboys crd to become active: %v", err)
					return false
				}
				return true
			}, wait.ForeverTestTimeout, time.Millisecond*100)

			// create virtual workspace rest configs
			rawConfig, err := server.RawConfig()
			require.NoError(t, err)
			virtualWorkspaceRawConfig := rawConfig.DeepCopy()
			virtualWorkspaceRawConfig.Clusters["kubelike"] = rawConfig.Clusters["system:admin"].DeepCopy()
			virtualWorkspaceRawConfig.Clusters["kubelike"].Server = rawConfig.Clusters["system:admin"].Server + "/services/syncer/" + kubelikeWorkspace.String() + "/kubelike"
			virtualWorkspaceRawConfig.Contexts["kubelike"] = rawConfig.Contexts["system:admin"].DeepCopy()
			virtualWorkspaceRawConfig.Contexts["kubelike"].Cluster = "kubelike"
			virtualWorkspaceRawConfig.Clusters["wildwest"] = rawConfig.Clusters["system:admin"].DeepCopy()
			virtualWorkspaceRawConfig.Clusters["wildwest"].Server = rawConfig.Clusters["system:admin"].Server + "/services/syncer/" + wildwestWorkspace.String() + "/" + wildwestWorkloadClusterName
			virtualWorkspaceRawConfig.Contexts["wildwest"] = rawConfig.Contexts["system:admin"].DeepCopy()
			virtualWorkspaceRawConfig.Contexts["wildwest"].Cluster = "wildwest"
			kubelikeVWConfig, err := clientcmd.NewNonInteractiveClientConfig(*virtualWorkspaceRawConfig, "kubelike", nil, nil).ClientConfig()
			require.NoError(t, err)
			wildwestVWConfig, err := clientcmd.NewNonInteractiveClientConfig(*virtualWorkspaceRawConfig, "wildwest", nil, nil).ClientConfig()
			require.NoError(t, err)

			t.Log("Starting test...")
			testCase.work(t, kubelikeVWConfig, wildwestVWConfig, kubelikeWorkspace, wildwestWorkspace)
		})
	}
}

type ByGroupVersion []*metav1.APIResourceList

func (a ByGroupVersion) Len() int           { return len(a) }
func (a ByGroupVersion) Swap(i, j int)      { a[i], a[j] = a[j], a[i] }
func (a ByGroupVersion) Less(i, j int) bool { return a[i].GroupVersion < a[j].GroupVersion }

type ByName []metav1.APIResource

func (a ByName) Len() int           { return len(a) }
func (a ByName) Swap(i, j int)      { a[i], a[j] = a[j], a[i] }
func (a ByName) Less(i, j int) bool { return a[i].Name < a[j].Name }

func sortAPIResourceList(list []*metav1.APIResourceList) []*metav1.APIResourceList {
	sort.Sort(ByGroupVersion(list))
	for _, resource := range list {
		sort.Sort(ByName(resource.APIResources))
	}
	return list
}
